/*
 * Copyright (C) 2012 Samsung Electronics
 * R. Chandrasekar <rcsekar at samsung.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <malloc.h>
#include <common.h>
#include <asm/io.h>
#include <i2c.h>
#include <i2s.h>
#include <asm/arch/i2s-regs.h>
#include <sound.h>
#include <asm/arch/sound.h>
#include <asm/arch/gpio.h>
#include <asm/arch/pinmux.h>
#include <wave.h>
#include "audio_file.h"

/* defines */
#define SOUND_BITS_IN_BYTE	8
#define GPIO_Y7			7
#define AUD_AMP			200

static struct i2stx_info g_i2stx_pri;
static struct sound_codec_info g_codec_info;

/*
 * get_sound_i2s_values gets fdt values for i2s parameters
 */
static void get_sound_i2s_values(struct i2stx_info *i2s)
{
	i2s->base_address = I2S_BASE_ADDRESS;
	i2s->audio_pll_clk = I2S_PLL_CLK;
	i2s->samplingrate = I2S_SAMPLING_RATE;
	i2s->bitspersample = I2S_BITS_PER_SAMPLE;
	i2s->channels = I2S_CHANNELS;
	i2s->rfs = I2S_RFS;
	i2s->bfs = I2S_BFS;
}

int audio_reg_init(void)
{
        /* Setting Clock Control Registers */
        writel(0xA12D0503, EPLL_CON0_REG);

        /* Setting I2S Registers */
        writel(0x9040c305, I2S_CON);
        writel(0xc0000400, I2S_MOD);
}

int sound_init(void)
{
	int ret;
	struct exynos5420_gpio_part1 *gpy7 = (struct exynos5420_gpio_part1 *)samsung_get_base_gpio_part1();
	struct i2stx_info *pi2s_tx = &g_i2stx_pri;
	struct sound_codec_info *pcodec_info = &g_codec_info;

	/* Get the I2S Values */
	get_sound_i2s_values(pi2s_tx);

	ret = i2s_tx_init(pi2s_tx);
	if (ret) {
		debug("%s: Failed to init i2c transmit: ret=%d\n", __func__, ret);
		return ret;
	}

	audio_reg_init();

	/* Setting GPY7 */
	s5p_gpio_direction_output(&gpy7->y7, GPIO_Y7, GPIO_OUTPUT);
	/* commenting this out. we should set this up in the board file instead */
	//s5p_gpio_set_pull(&gpy7->y7, GPIO_Y7, GPIO_PULL_DOWN);

	return ret;
}

/*
 * Fills buffer with boot audio
 */
static void sound_prepare_buffer(unsigned short *data, int size)
{
	const unsigned short amplitude = AUD_AMP; /* between 1 and 32767 */
	int i;

	for (i = 0; i < size; i++)
		*data++ = amplitude * ((unsigned short)amp[i]);
}

int sound_play(uint32_t msec, uint32_t frequency)
{
	unsigned int *data;
	unsigned long data_size;
	unsigned int ret = 0;
	struct i2s_reg *i2s_reg;
	struct exynos5420_gpio_part1 *gpy7 = (struct exynos5420_gpio_part1 *)samsung_get_base_gpio_part1();

	/*Buffer length computation */
	data_size = g_i2stx_pri.samplingrate * g_i2stx_pri.channels;
	data_size *= (g_i2stx_pri.bitspersample / SOUND_BITS_IN_BYTE);

	data = malloc(data_size);
	data_size = ARRAY_SIZE(amp);

	if (data == NULL) {
		debug("%s: malloc failed\n", __func__);
		return -1;
	}

	sound_prepare_buffer((unsigned short *)data,
	ARRAY_SIZE(amp));
        i2s_reg = (struct i2s_reg *)g_i2stx_pri.base_address;
	ret = i2s_transfer_tx_data(&g_i2stx_pri, data, data_size);

	s5p_gpio_direction_output(&gpy7->y7, GPIO_Y7, GPIO_INPUT);
	free(data);

	return ret;
}

int play_audio (short * data, unsigned int nsamples, int channel)
{
	unsigned int i;
	struct i2s_reg *i2s_reg;
	struct exynos5420_gpio_part1 *gpy7;
	unsigned char buf[4];
	int pcm;

	gpy7 =  (struct exynos5420_gpio_part1 *)samsung_get_base_gpio_part1();
	i2s_reg = (struct i2s_reg *)g_i2stx_pri.base_address;

	if (nsamples < FIFO_LENGTH)
		return 0;

	/* Is it mono/stereo? */
	if ( !is_channel_supported(channel) )
		return 0;

	for (i = 0; i < FIFO_LENGTH; i++) {
		if (channel == MONO) {
			memcpy(&buf[0], data, 2); /* Left  Channel */
			memcpy(&buf[2], data, 2); /* Right Channel: make it stereo  */
			data++;
		} else	if (channel == STEREO) {
			memcpy(buf, data, 4); /* one frame: 16 bits left and 16 bits right */
			data++;
			data++;
		}

		memcpy(&pcm, buf, sizeof(pcm));
		writel(pcm, &i2s_reg->txd);

	}

	nsamples -= FIFO_LENGTH;
	i2s_txctrl(i2s_reg, I2S_TX_ON);

	while (nsamples > 0) {
		if (!(CON_TXFIFO_FULL & (readl(&i2s_reg->con)))) {
			if (channel == MONO) {
				/* convert mono to stereo */
				memcpy(&buf[0], data, 2); /* Left  Channel */
				memcpy(&buf[2], data, 2); /* Right Channel: make it stereo */
				data++;
			} else	if (channel == STEREO) {
				memcpy(buf, data, 4); /* one frame: 16 bits left and 16 bits right */
				data++;
				data++;
			}

			memcpy(&pcm, buf, sizeof(pcm));
			writel(pcm, &i2s_reg->txd);
			nsamples--;
		}
	}

	i2s_txctrl(i2s_reg, I2S_TX_OFF);
	s5p_gpio_direction_output(&gpy7->y7, GPIO_Y7, GPIO_INPUT);

	return 1;
}
